import os
import torch
import os.path as osp
from dig.xgraph.models import *
from dig.xgraph.dataset import SynGraphDataset
from torch_geometric.data import DataLoader
from torch_geometric.data import Data, InMemoryDataset, download_url, extract_zip


device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


def index_to_mask(index, size):
    mask = torch.zeros(size, dtype=torch.bool, device=index.device)
    mask[index] = 1
    return mask


def split_dataset(dataset):
    indices = []
    num_classes = 4
    train_percent = 0.7
    for i in range(num_classes):
        index = (dataset.data.y == i).nonzero().view(-1)
        index = index[torch.randperm(index.size(0))]
        indices.append(index)

    train_index = torch.cat([i[:int(len(i) * train_percent)] for i in indices], dim=0)

    rest_index = torch.cat([i[int(len(i) * train_percent):] for i in indices], dim=0)
    rest_index = rest_index[torch.randperm(rest_index.size(0))]

    dataset.data.train_mask = index_to_mask(train_index, size=dataset.data.num_nodes)
    dataset.data.val_mask = index_to_mask(rest_index[:len(rest_index) // 2], size=dataset.data.num_nodes)
    dataset.data.test_mask = index_to_mask(rest_index[len(rest_index) // 2:], size=dataset.data.num_nodes)

    dataset.data, dataset.slices = dataset.collate([dataset.data])

    return dataset


dataset = SynGraphDataset('./datasets', 'ba_2motifs')
dataset.data.x = dataset.data.x.to(torch.float32)
dataset.data.x = dataset.data.x[:, :1]
# dataset.data.y = dataset.data.y[:, 2]
dim_node = dataset.num_node_features
dim_edge = dataset.num_edge_features
# num_targets = dataset.num_classes
num_classes = dataset.num_classes

# splitted_dataset = split_dataset(dataset)
# splitted_dataset.data.mask = splitted_dataset.data.test_mask
# splitted_dataset.slices['mask'] = splitted_dataset.slices['train_mask']
# dataloader = DataLoader(splitted_dataset, batch_size=1, shuffle=False)


def check_checkpoints(root='./'):
    if osp.exists(osp.join(root, 'checkpoints')):
        return
    url = ('https://github.com/divelab/DIG_storage/raw/main/xgraph/checkpoints.zip')
    path = download_url(url, root)
    extract_zip(path, root)
    os.unlink(path)


model = GCN_2l_mask(model_level='graph', dim_node=dim_node, dim_hidden=300, num_classes=num_classes)
model.to(device)
check_checkpoints()
ckpt_path = osp.join('checkpoints', 'ba_shapes', 'GCN_2l', '0', 'GCN_2l_best.ckpt')
# model.load_state_dict(torch.load(ckpt_path)['state_dict'])


from dig.xgraph.method import PGExplainer
explainer = PGExplainer(model, in_channels=600, device=device, explain_graph=True)


# --- Create data collector and explanation processor ---
from dig.xgraph.evaluation import XCollector
x_collector = XCollector()

data = dataset[0]

from dig.xgraph.method.pgexplainer import PlotUtils
plotutils = PlotUtils(dataset_name='ba_2motifs')


# graph classification
max_nodes = 5
index = -1
for i, data in enumerate(dataset):
    index += 1
    print(f'explain graph {index}')
    data.to(device)

    if torch.isnan(data.y[0].squeeze()):
        continue

    logits = model(data.x, data.edge_index)
    prediction = logits.argmax(-1).item()

    _, masks, related_preds = \
        explainer(data.x, data.edge_index, max_nodes=max_nodes, y=data.y)
    explainer.visualization(data, edge_mask=masks[0], top_k=5, plot_utils=plotutils)
    x_collector.collect_data(masks, related_preds)
    if index >= 20:
        break

print(f'Fidelity: {x_collector.fidelity:.4f}\n'
      f'Sparsity: {x_collector.sparsity:.4f}')


# node classification
# # explainer.train_explanation_network(splitted_dataset)
# torch.save(explainer.state_dict(), 'tmp.pt')
# state_dict = torch.load('tmp.pt')
# explainer.load_state_dict(state_dict)
#
# # --- Create data collector and explanation processor ---
# from dig.xgraph.evaluation import XCollector, ExplanationProcessor
# x_collector = XCollector()
#
# random_seed = 0
# import random
# import numpy as np
# random.seed(random_seed)
# np.random.seed(random_seed)
# torch.manual_seed(random_seed)
# torch.cuda.manual_seed_all(random_seed)
#
# node_indices = torch.where(dataset[0].test_mask * dataset[0].y != 0)[0].tolist()
# from dig.xgraph.method.pgexplainer import PlotUtils
# plotutils = PlotUtils(dataset_name='ba_shapes')
# data = dataset[0]
# node_idx = node_indices[6]
# walks, masks, related_preds = \
#     explainer(data.x, data.edge_index, node_idx=node_idx, y=data.y, top_k=5)
#
# explainer.visualization(data, edge_mask=masks[0], top_k=5, plot_utils=plotutils, node_idx=node_idx)
#
# index = -1
# for i, data in enumerate(dataloader):
#     for j, node_idx in enumerate(node_indices):
#         index += 1
#         print(f'explain graph {i} node {node_idx}')
#         data.to(device)
#
#         walks, masks, related_preds = \
#             explainer(data.x, data.edge_index, node_idx=node_idx, y=data.y, top_k=5)
#
#         x_collector.collect_data(masks, related_preds)
#
#         if index >= 99:
#             break
#
#     if index >= 99:
#         break
#
#
# print(f'Fidelity: {x_collector.fidelity:.4f}\n'
#       f'Sparsity: {x_collector.sparsity:.4f}')
